/* file : bloom-filter.ts
MIT License

Copyright (c) 2017 Thomas Minier & Arnaud Grall

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/
'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
var formulas_1 = require("./formulas");
var utils_1 = require("./utils");
var encoding_1 = require("./encoding");
/**
 * A Bloom filter is a space-efficient probabilistic data structure, conceived by Burton Howard Bloom in 1970,
 * that is used to test whether an element is a member of a set. False positive matches are possible, but false negatives are not.
 *
 * Reference: Bloom, B. H. (1970). Space/time trade-offs in hash coding with allowable errors. Communications of the ACM, 13(7), 422-426.
 * @see {@link http://crystal.uta.edu/~mcguigan/cse6350/papers/Bloom.pdf} for more details about classic Bloom Filters.
 * @author Thomas Minier
 * @author Arnaud Grall
 */
var BloomFilter = /** @class */ (function () {
    /**
     * Constructor
     * @param size - The number of cells.
     * @param nbHashes - The number of hash functions used
     */
    function BloomFilter(size, nbHashes, seed) {
        if (seed === void 0) { seed = 0x1234567890; }
        var positiveIntSeed = Math.abs(Math.ceil(seed));
        this._seed = positiveIntSeed;
        this._size = size;
        this._nbHashes = nbHashes;
        this._filter = new Uint8Array(utils_1.getUint8ArrayLength(size));
        this._length = 0;
    }
    /**
     * Create an optimal bloom filter providing the maximum of elements stored and the error rate desired
     * @param  items      - The maximum nuber of item to store
     * @param  errorRate  - The error rate desired for a maximum of items inserted
     * @return A new {@link BloomFilter}
     */
    BloomFilter.create = function (nbItems, errorRate, seed) {
        if (seed === void 0) { seed = 0x1234567890; }
        var size = formulas_1.optimalFilterSize(nbItems, errorRate);
        var hashes = formulas_1.optimalHashes(size, nbItems);
        return new BloomFilter(size, hashes, seed);
    };
    /**
     * Build a new Bloom Filter from an existing iterable with a fixed error rate
     * @param items - The iterable used to populate the filter
     * @param errorRate - The error rate, i.e. 'false positive' rate, targetted by the filter
     * @return A new Bloom Filter filled with the iterable's elements
     * @example
     * // create a filter with a false positive rate of 0.1
     * const filter = BloomFilter.from(['alice', 'bob', 'carl'], 0.1);
     */
    BloomFilter.from = function (items, errorRate, seed) {
        if (seed === void 0) { seed = 0x1234567890; }
        var array = Array.from(items);
        var filter = BloomFilter.create(array.length, errorRate, seed);
        array.forEach(function (element) { return filter.add(element); });
        return filter;
    };
    /**
     * Generate a bloom filter from a binary export.
     * @param  binaryBloomFilter - The bloom filter as a Uint8Array
     * @return Bloom filter generated from the exported binary filter
     * @author Kolja Blauhut
     */
    BloomFilter.import = function (binaryBloomFilter) {
        var seedArray = binaryBloomFilter.slice(0, 8);
        var nbHashesArray = binaryBloomFilter.slice(8, 16);
        var lengthArray = binaryBloomFilter.slice(16, 24);
        var sizeArray = binaryBloomFilter.slice(24, 32);
        var filterArray = binaryBloomFilter.slice(32, binaryBloomFilter.length);
        var bloomFilter = new BloomFilter(encoding_1.uint8ArrayToInt64(sizeArray), encoding_1.uint8ArrayToInt64(nbHashesArray), encoding_1.uint8ArrayToInt64(seedArray));
        bloomFilter._length = encoding_1.uint8ArrayToInt64(lengthArray);
        bloomFilter._filter = filterArray;
        return bloomFilter;
    };
    Object.defineProperty(BloomFilter.prototype, "size", {
        /**
         * Get the optimal size of the filter
         * @return The size of the filter
         */
        get: function () {
            return this._size;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(BloomFilter.prototype, "length", {
        /**
         * Get the number of elements currently in the filter
         * @return The filter length
         */
        get: function () {
            return this._length;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(BloomFilter.prototype, "seed", {
        /**
         * Get the seed of the filter
         * @return The filter seed
         */
        get: function () {
            return this._seed;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(BloomFilter.prototype, "nbHashes", {
        /**
         * Get the hash count of the filter
         * @return The filter hash count
         */
        get: function () {
            return this._nbHashes;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Add an element to the filter
     * @param element - The element to add
     * @example
     * const filter = new BloomFilter(15, 0.1);
     * filter.add('foo');
     */
    BloomFilter.prototype.add = function (element) {
        var indexes = utils_1.getDistinctIndices(element, this._size, this._nbHashes, this._seed);
        for (var i = 0; i < indexes.length; i++) {
            var indexToEdit = utils_1.getByteIndexInArray(indexes[i]);
            var byteToEdit = this._filter[indexToEdit];
            var editedByte = utils_1.setBitInByte(utils_1.getBitIndex(indexes[i]), byteToEdit);
            this._filter[indexToEdit] = editedByte;
        }
        this._length++;
    };
    /**
     * Test an element for membership
     * @param element - The element to look for in the filter
     * @return False if the element is definitively not in the filter, True is the element might be in the filter
     * @example
     * const filter = new BloomFilter(15, 0.1);
     * filter.add('foo');
     * console.log(filter.has('foo')); // output: true
     * console.log(filter.has('bar')); // output: false
     */
    BloomFilter.prototype.has = function (element) {
        var indexes = utils_1.getDistinctIndices(element, this._size, this._nbHashes, this._seed);
        for (var i = 0; i < indexes.length; i++) {
            if (!utils_1.getBitAtIndex(this._filter, indexes[i])) {
                return false;
            }
        }
        return true;
    };
    /**
     * Get the current false positive rate (or error rate) of the filter
     * @return The current false positive rate of the filter
     * @example
     * const filter = new BloomFilter(15, 0.1);
     * console.log(filter.rate()); // output: something around 0.1
     */
    BloomFilter.prototype.rate = function () {
        return Math.pow(1 - Math.exp((-this._nbHashes * this._length) / this._size), this._nbHashes);
    };
    /**
     * Check if another Bloom Filter is equal to this one
     * @param  filter - The filter to compare to this one
     * @return True if they are equal, false otherwise
     */
    BloomFilter.prototype.equals = function (other) {
        if (this._size !== other._size ||
            this._nbHashes !== other._nbHashes ||
            this._length !== other._length) {
            return false;
        }
        return this._filter.every(function (value, index) { return other._filter[index] === value; });
    };
    /**
     * Generate a binary export for the bloom filter.
     * @return Binary Unit8Array export of the bloom filter
     * @author Kolja Blauhut
     */
    BloomFilter.prototype.export = function () {
        var exportArray = new Uint8Array(this._filter.length + 4 * 8); // Filter length + 4 number parameters
        exportArray.set(encoding_1.int64ToUint8Array(this.seed), 0);
        exportArray.set(encoding_1.int64ToUint8Array(this.nbHashes), 8);
        exportArray.set(encoding_1.int64ToUint8Array(this.length), 16);
        exportArray.set(encoding_1.int64ToUint8Array(this.size), 24);
        var exportArrayStartIndex = 32;
        for (var index = 0; index < this._filter.length; index += 1) {
            exportArray[index + exportArrayStartIndex] = this._filter[index];
        }
        return exportArray;
    };
    return BloomFilter;
}());
exports.default = BloomFilter;
