/* file : utils.ts
MIT License

Copyright (c) 2017-2020 Thomas Minier & Arnaud Grall

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/
'use strict';
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getBitAtIndex = exports.setBitInByte = exports.getBitIndex = exports.getByteIndexInArray = exports.getUint8ArrayLength = exports.getDefaultSeed = exports.getDistinctIndices = exports.doubleHashing = exports.hashTwice = void 0;
var xxhashjs_1 = __importDefault(require("xxhashjs"));
/**
 * (64-bits only) Hash a value into two values (in hex or integer format)
 * @param  value - The value to hash
 * @param  asInt - (optional) If True, the values will be returned as an integer. Otherwise, as hexadecimal values.
 * @param seed the seed used for hashing
 * @return The results of the hash functions applied to the value (in hex or integer)
 * @memberof Utils
 * @author Arnaud Grall & Thomas Minier
 */
function hashTwice(value, asInt, seed) {
    if (asInt === undefined) {
        asInt = false;
    }
    if (seed === undefined) {
        seed = getDefaultSeed();
    }
    var f = xxhashjs_1.default.h64(value, seed + 1);
    var l = xxhashjs_1.default.h64(value, seed + 2);
    if (asInt) {
        return {
            first: f.toNumber(),
            second: l.toNumber(),
        };
    }
    else {
        var one = f.toString(16);
        if (one.length < 16) {
            one = '0'.repeat(16 - one.length) + one;
        }
        var two = l.toString(16);
        if (two.length < 16) {
            two = '0'.repeat(16 - two.length) + two;
        }
        return {
            first: Number(one),
            second: Number(two),
        };
    }
}
exports.hashTwice = hashTwice;
/**
 * Apply Double Hashing to produce a n-hash
 *
 * This implementation used directly the value produced by the two hash functions instead of the functions themselves.
 * @see {@link http://citeseer.ist.psu.edu/viewdoc/download;jsessionid=4060353E67A356EF9528D2C57C064F5A?doi=10.1.1.152.579&rep=rep1&type=pdf} for more details about double hashing.
 * @param  n - The indice of the hash function we want to produce
 * @param  hashA - The result of the first hash function applied to a value.
 * @param  hashB - The result of the second hash function applied to a value.
 * @param  size - The size of the datastructures associated to the hash context (ex: the size of a Bloom Filter)
 * @return The result of hash_n applied to a value.
 * @memberof Utils
 * @author Thomas Minier
 */
function doubleHashing(n, hashA, hashB, size) {
    return Math.abs((hashA + n * hashB) % size);
}
exports.doubleHashing = doubleHashing;
/**
 * Generate a set of distinct indexes on interval [0, size) using the double hashing technique
 * @param  element  - The element to hash
 * @param  size     - the range on which we can generate an index [0, size) = size
 * @param  number   - The number of indexes desired
 * @param  seed     - The seed used
 * @return A array of indexes
 * @author Arnaud Grall
 */
function getDistinctIndices(element, size, number, seed) {
    if (seed === undefined) {
        seed = getDefaultSeed();
    }
    function getDistinctIndicesBis(n, elem, size, count, indexes) {
        if (indexes === void 0) { indexes = []; }
        if (indexes.length === count) {
            return indexes;
        }
        else {
            var hashes = hashTwice(elem, true, seed + (size % n));
            var ind = doubleHashing(n, hashes.first, hashes.second, size);
            if (indexes.includes(ind)) {
                // console.log('generate index: %d for %s', ind, elem)
                return getDistinctIndicesBis(n + 1, elem, size, count, indexes);
            }
            else {
                // console.log('already found: %d for %s', ind, elem)
                indexes.push(ind);
                return getDistinctIndicesBis(n + 1, elem, size, count, indexes);
            }
        }
    }
    return getDistinctIndicesBis(1, element, size, number);
}
exports.getDistinctIndices = getDistinctIndices;
/**
 * Return the default seed used in the package
 * @return A ssed as a floating point number
 * @author Arnaud Grall
 */
function getDefaultSeed() {
    return 0x1234567890;
}
exports.getDefaultSeed = getDefaultSeed;
/**
 * Return the amount of bytes needed to fit the input bits
 * @return Length of Unit8Array to use
 * @param bitCount    - amount of bits the filter uses
 */
function getUint8ArrayLength(bitCount) {
    var remainder = bitCount % 8;
    var bitFill = 8 - remainder;
    return (bitCount + bitFill) / 8;
}
exports.getUint8ArrayLength = getUint8ArrayLength;
/**
 * Return the index of the byte to be edited within the array
 * @return Array index of the byte to be edited
 * @param bitIndex    - index of the bit to be set
 * @author Kolja Blauhut
 */
function getByteIndexInArray(bitIndex) {
    return Math.floor(bitIndex / 8);
}
exports.getByteIndexInArray = getByteIndexInArray;
/**
 * Return the index of the bit in the byte to edit
 * @return Array index of the byte to be edited
 * @param bitIndex    - index of the bit to be set
 * @author Kolja Blauhut
 */
function getBitIndex(bitIndex) {
    return bitIndex % 8;
}
exports.getBitIndex = getBitIndex;
/**
 * Set a certain bit in the byte to 1
 * @return Edited byte
 * @param indexInByte     - Index of the bit in the byte to be set
 * @param byte            - Current byte
 * @author Kolja Blauhut
 */
function setBitInByte(indexInByte, byte) {
    var byteOR = 1 << indexInByte;
    return byte | byteOR;
}
exports.setBitInByte = setBitInByte;
/**
 * Returns a bit at a given index
 * @return Bit 1 | 0
 * @param array     - Uint8Array containing bloom filter
 * @param bitIndex  - Index of bit to read
 * @author Kolja Blauhut
 */
function getBitAtIndex(array, bitIndex) {
    var byte = array[getByteIndexInArray(bitIndex)];
    var indexInByte = getBitIndex(bitIndex);
    var byteAND = setBitInByte(indexInByte, 0);
    return ((byte & byteAND) >> indexInByte);
}
exports.getBitAtIndex = getBitAtIndex;
